/*****************************************************************************
 * vlf.c: simple MPEG stream server
 *****************************************************************************
 * Copyright (C) 2001 VideoLAN
 *
 * Authors:
 * Samuel Hocevar <sam@zoy.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

#include <fcntl.h>                                                   /* open */
#include <stdio.h>                                                 /* printf */
#include <getopt.h>                                              /* getopt() */
#include <stdlib.h>                                                  /* atoi */
#include <sys/stat.h>                                               /* fstat */
#include <sys/time.h>                                             /* timeval */
#include <unistd.h>                                    /* usleep (not POSIX) */
#include <netinet/in.h>                                             /* ntohl */
#include <sys/ioctl.h>                                              /* ioctl */
#include <netdb.h>                                              /* inet_ntoa */
#include <net/if.h>                                        /* for netutils.h */
#include <string.h>                                                 /* errno */
#include <errno.h>                                                  /* errno */

#include "types.h"                                                     /* u8 */
#include "netutils.h"                /* netutils (get the broadcast address) */

#define TS_PACKET_SIZE 188
#define TS_IN_UDP 7
#define LOOP_SIZE 20

#define COPYRIGHT \
"VideoLAN Network Tester - (c)2001 VideoLAN\n\n"

#define VERSION \
COPYRIGHT \
"This program comes with NO WARRANTY, to the extent permitted by law.\n" \
"You may redistribute it under the terms of the GNU General Public " \
"License;\n" \
"see the file named COPYING for details.\n" \
"Written by the VideoLAN team at Ecole Centrale, Paris.\n"

#define HELP \
"Usage: %s [OPTION]\n" \
COPYRIGHT \
"\n" \
"Options:\n" \
"  -d host[:port]           target address (sets server mode)\n" \
"  -c                       client mode\n" \
"  -r rate                  rate (default 128kbit/s)\n" \
"\n" \
"  -h, --help               display this help and exit\n" \
"  -v, --version            output version information and exit\n"

/* Long options */
static const struct option longopts[] =
{
    /*  name,               has_arg,    flag,   val */
    {   "version",          0,          0,      'v' },
    {   "help",             0,          0,      'h' },
    {   0,                  0,          0,      0 }
};

/* Short options */
static const char *shortopts = "d:r:chv";

typedef struct compteur_s
{
    u8 a1;
    u8 a2;
    u8 a3;
    u8 a4;

    u8 a5;
    u8 a6;
    u8 a7;
    u8 a8;
} compteur_t;

/* Prototypes */
static void main_server( char *client, int kbit_per_second );
static void main_client( void );
static int open_socket( char *host, struct sockaddr_in *remote_addr );
static void usage( char *prog );
static void version( char *prog );

/******************************************************************************
 * GetTime : returns the current time in microseconds
 ******************************************************************************/
s64 GetTime(void)
{
    struct timeval tmp;

    gettimeofday(&tmp, NULL);
    return( (s64)tmp.tv_sec * 1000000 + (s64)tmp.tv_usec );
}

/******************************************************************************
 * msleep : sleeps 'delay' microseconds
 ******************************************************************************/
void msleep( s64 delay )
{
    struct timeval tv_delay;

    tv_delay.tv_sec = delay / 1000000;
    tv_delay.tv_usec = delay % 1000000;
    /* select() return value should be tested, since several possible errors
     * can occur. However, they should only happen in very particular occasions
     * (i.e. when a signal is sent to the thread, or when memory is full), and
     * can be ignored. */
    select( 0, NULL, NULL, NULL, &tv_delay );
}

/******************************************************************************
 * mdate: get date
 ******************************************************************************/
s64 mdate( void )
{
    struct timeval tv_date;

    /* gettimeofday() could return an error, and should be tested. However, the
     * only possible error, according to 'man', is EFAULT, which can not happen
     * here, since tv is a local variable. */
    gettimeofday( &tv_date, NULL );
    return( (s64) tv_date.tv_sec * 1000000 + (s64) tv_date.tv_usec );
}

/******************************************************************************
 * main : Main mini-server program
 ******************************************************************************/
int main(int argc, char **argv)
{
    char *client = NULL;
    int kbit_per_second = 128;
    int opt, server = 0;

    /* Argument checking */
    /* optional */
    opterr = 0; /* force getopt not to generate an error message for '?' */

    /* while((opt = getopt(argc, argv, shortopts)) != EOF) */
    while((opt = getopt_long( argc, argv, shortopts, longopts, 0)) != EOF)
    {
        switch(opt)
        {
        case 'd':
            client = optarg;
            break;
        case 'c':
            server = 1;
            break;
        case 'r':
            kbit_per_second = atoi(optarg);
            break;
        case 'h' :
            usage( *argv );
            exit( 0 );
            break;
        case 'v' :
            version( *argv );
            exit( 0 );
            break;
        case '?':
            fprintf( stderr, "E: invalid option or missing parameter `%c'\n",
                     optopt );
            usage( *argv );
            exit( 1 );
            break;
        }
    }

    /* Copyright information */
    fprintf( stderr, COPYRIGHT );
    
    /* Don't use output buffer to have a real-time display */
    setvbuf( stdout, NULL, _IONBF ,0 );

    if( client != NULL )
    {
        if( server )
        {
            fprintf( stderr, "E: you specified both client mode"
                               " and a destination address\n" );
        }
        else /* server == 0 */
        {
            main_server( client, kbit_per_second );
        }
    }
    else /* client == NULL */
    {
        if( server )
        {
            main_client( );
        }
        else /* server == 0 */
        {
            fprintf( stderr, "E: must specify client mode"
                               " or a destination address\n" );
        }
    }

    return( 0 );
}

static void main_client( void )
{
    char *packet;
    int fd, opt, received;
    struct sockaddr_in local_addr;
    s64 count, oldcount = 0;
    s64 time, oldtime = 0;
    int packets = 0;

    packet = malloc( sizeof(u8) * TS_PACKET_SIZE * TS_IN_UDP );

    fd = socket( AF_INET, SOCK_DGRAM, 0 );
    if( fd == -1 )
    {
        fprintf( stderr, "E: can't create socket\n" );
    }

    if( setsockopt( fd, SOL_SOCKET, SO_REUSEADDR,
                    (void*) &opt, sizeof( opt ) ) == -1 )
    {
        fprintf( stderr, "E: can't configure socket (SO_REUSEADDR)\n" );
        close(fd);
    }

    opt = 0x80000;
    if( setsockopt( fd, SOL_SOCKET, SO_RCVBUF,
                    (void*) &opt, sizeof( opt ) ) == -1 )
    {
        fprintf( stderr, "E: can't configure socket (SO_RCVBUF)\n" );
        close(fd);
    }

    local_addr.sin_family      = AF_INET;
    local_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    local_addr.sin_port        = htons(DEFAULT_PORT);

    bind( fd, (struct sockaddr *)&local_addr, sizeof( local_addr ) );

    fprintf(stderr,"I: waiting for packets\n");
    while(1)
    {
        int i;

        for( i = 0; i < LOOP_SIZE ; i++ )
        {
            received = recv( fd, packet, TS_PACKET_SIZE * TS_IN_UDP, 0 );

            count  = ((compteur_t *)packet)->a5;
            count |= ((compteur_t *)packet)->a6 << 8;
            count |= ((compteur_t *)packet)->a7 << 16;
            count |= ((compteur_t *)packet)->a8 << 24;

            if( count != oldcount + 1 && oldcount != 0 )
            {
                fprintf(stderr,"E: mismatch, %lli then %lli\n", oldcount, count);
            }

            oldcount = count;
        }

        packets += LOOP_SIZE;
        time = mdate();

        if( time > oldtime + 1000000 )
        {
            fprintf(stderr, "I: read %i packets\n", packets );
            oldtime = time;
            packets = 0;
        }
    }
}

static void main_server( char *client, int kbit_per_second )
{
    char *packet;
    int fd, count;
    s64 old, new, average, delay, ratio = (s64)(TS_IN_UDP * TS_PACKET_SIZE) * (s64)(1000000 * 8 / 1024);
    struct compteur_s compteur;
    struct sockaddr_in remote_addr;

    /*
     * Main network loop
     */
    packet = malloc( sizeof(u8) * TS_PACKET_SIZE * TS_IN_UDP );
    memset( packet, 0x00, TS_PACKET_SIZE * TS_IN_UDP );

    memset( &compteur, 0x00, sizeof( compteur ) );
    compteur.a1 = 'X';
    compteur.a2 = 'x';
    compteur.a3 = 'O';
    compteur.a4 = 'o';

    fd = open_socket( client, &remote_addr );
    count = 0;
    average = 0;
    
    fprintf( stderr, "I: started\n" );
    while(1)
    {
        int i;

        delay = 0;
        old = mdate();
        new = old + LOOP_SIZE * ratio / kbit_per_second;

        for( i = 0; i < LOOP_SIZE ; i++ )
        {
            /* Copy the structure so that we know which packet it is */
            compteur.a5 = (count & 0x000000ff) >> 0;
            compteur.a6 = (count & 0x0000ff00) >> 8;
            compteur.a7 = (count & 0x00ff0000) >> 16;
            compteur.a8 = (count & 0xff000000) >> 24;
            memcpy( packet, &compteur, sizeof( compteur ) );

            /* Send the UDP packet over the network */
            sendto( fd, packet, TS_IN_UDP * TS_PACKET_SIZE, 0,
                    (struct sockaddr *) &remote_addr, sizeof(remote_addr) );
            count++;

            if( mdate() < new && ratio / kbit_per_second > 2000 )
            {
                msleep( ratio / kbit_per_second - 2000 );
            }
        }

        new = mdate();

        if( new < old + LOOP_SIZE * ratio / kbit_per_second - 2000 )
        {
            msleep( old - new + LOOP_SIZE * ratio / kbit_per_second - 2000 );
        }

        new = mdate();

        if( average )
        {
            average *= 15;
            average += (ratio * LOOP_SIZE) / (s64)( new - old );
            average /= 16;
        }
        else
        {
            average = (ratio * LOOP_SIZE) / (s64)( new - old );
        }

        fprintf( stderr, "I: speed %lli kbit/s    \t(average %lli kbit/s)   
",
                 (ratio * LOOP_SIZE) / (s64)( new - old ), average );
    }
}

/******************************************************************************
 * open_socket
 ******************************************************************************/
static int open_socket( char *host, struct sockaddr_in *remote_addr )
{
    int socket_fd, sockopt;
    struct sockaddr_in local_addr;
    
    /* open a socket for UDP/IP */
    if( (socket_fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0 )
    {
        fprintf( stderr, "E: socket() error\n" );
    }

    BuildInetAddr( &local_addr, NULL );
    if( host != NULL )
    {
        BuildInetAddr( remote_addr, host );
    }
    else
    {
        int i_if_nb;
        net_descr_t net_descr;

        /* Get informations about the net configuration (interface ...)
         * to retrieve the broadcast address we need */
        ReadNetConf( socket_fd, &net_descr );
        i_if_nb = net_descr.i_if_number;
        if( i_if_nb == 0 )
        {
            fprintf( stderr, "E: Fatal error: no interface detected\n" );
            close( socket_fd );
            return( -1 );
        }
        fprintf( stderr, "I: Broadcast on interface %s\n",
                 net_descr.a_if[i_if_nb-1].psz_ifname );
        *remote_addr = net_descr.a_if[i_if_nb-1].sa_bcast_addr;
        free( net_descr.a_if );

        fprintf( stderr,"I: Sending to default port %i\n", DEFAULT_PORT );
        remote_addr->sin_port = htons( DEFAULT_PORT );
    }

    sockopt = 1;
    if( setsockopt( socket_fd, SOL_SOCKET, SO_BROADCAST,
                    &sockopt, sizeof(sockopt) ) < 0 )
    {
        fprintf( stderr, "E: SO_BROADCAST setsockopt error\n" );
    }
    
    if( bind(socket_fd, (struct sockaddr*)&local_addr, sizeof(local_addr)) < 0 )
    {
        fprintf( stderr, "E: bind() error\n" );
    }
    
    return socket_fd;
}

/******************************************************************************
 * usage : Show an help message and exit
 ******************************************************************************/
static void usage( char *prog )
{
    fprintf( stderr, HELP, prog );
}

/******************************************************************************
 * version : Show a version message and exit
 ******************************************************************************/
static void version( char *prog )
{
    fprintf( stderr, VERSION );
}

